import logging
import subprocess
import typing as t
import webbrowser

import dvc.api
import typer
import uvicorn
from dvc.stage.serialize import to_single_stage_lockfile

from paraffin.db import complete_job, get_job, save_graph_to_db, set_job_deps_lock
from paraffin.ui.app import app as webapp
from paraffin.utils import get_changed_stages, get_custom_queue, get_stage_graph

log = logging.getLogger(__name__)

app = typer.Typer()


@app.command()
def ui(port: int = 8000):
    """Start the Paraffin web UI."""
    webbrowser.open(f"http://localhost:{port}")
    uvicorn.run(webapp, host="0.0.0.0", port=port)


@app.command()
def worker(
    queues: str = typer.Option(
        "default",
        "--queues",
        "-q",
        envvar="PARAFFIN_QUEUES",
        help="Comma separated list of queues to listen on.",
    ),
):
    """Start a Celery worker."""
    queues = queues.split(",")
    print(queues)
    while True:
        job = get_job(queues=queues)
        if job is None:
            # TODO: timeout
            print("No job found.")
            return

        fs = dvc.api.DVCFileSystem(url=None, rev=None)
        with fs.repo.lock:
            stage = fs.repo.stage.collect(job["name"])[0]
            stage.save(allow_missing=True)
        stage_lock = to_single_stage_lockfile(stage, with_files=True)
        set_job_deps_lock(job["id"], stage_lock)

        try:
            subprocess.check_call(f"dvc repro -s {job['name']}", shell=True)
        except Exception as e:
            log.error(f"Failed to run job: {e}")
            complete_job(job["id"], status="failed")
        else:
            # get the stage_lock
            fs = dvc.api.DVCFileSystem(url=None, rev=None)
            with fs.repo.lock:
                stage = fs.repo.stage.collect(job["name"])[0]
                stage.save()
            stage_lock = to_single_stage_lockfile(stage, with_files=True)
            complete_job(job["id"], status="completed", lock=stage_lock)


@app.command()
def submit(
    names: t.Optional[list[str]] = typer.Argument(
        None, help="Stage names to run. If not specified, run all stages."
    ),
    verbose: bool = typer.Option(False, help="Verbose output."),
    check: bool = typer.Option(True, help="Check if stages are changed."),
):
    """Run DVC stages in parallel."""
    if verbose:
        logging.basicConfig(level=logging.DEBUG)

    log.debug("Getting stage graph")
    graph = get_stage_graph(names=names, glob=True)
    if check:
        changed = get_changed_stages(graph)
    else:
        changed = [node.name for node in graph.nodes]
    print(f"Changed stages: {changed}")
    print(f"Submitting {graph}")
    custom_queues = get_custom_queue()
    save_graph_to_db(graph, queues=custom_queues, changed=changed)



from pathlib import Path

from fastapi import FastAPI
from fastapi.responses import RedirectResponse
from fastapi.staticfiles import StaticFiles

from paraffin.db import db_to_graph
from paraffin.utils import build_elk_hierarchy

FILE = Path(__file__)

app = FastAPI()

app.mount("/ui", StaticFiles(directory=FILE.parent.parent / "static"), name="ui")
app.mount(
    "/assets",
    StaticFiles(directory=FILE.parent.parent / "static" / "assets"),
    name="assets",
)


@app.get("/")
def read_root():
    return RedirectResponse(url="/ui/index.html")


@app.get("/api/v1/graph")
def read_graph():
    g = db_to_graph()
    return build_elk_hierarchy(g)


@app.get("/api/v1/spawn")
def spawn():
    pass  # run paraffin worker



import fnmatch
import json
from typing import List, Optional

import networkx as nx
from dvc.stage.cache import _get_cache_hash
from sqlmodel import Field, Relationship, Session, SQLModel, create_engine, select


class JobDependency(SQLModel, table=True):
    parent_id: Optional[int] = Field(foreign_key="job.id", primary_key=True)
    child_id: Optional[int] = Field(foreign_key="job.id", primary_key=True)


class Job(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    name: str
    cmd: str  # Command to execute
    status: str = "pending"  # pending, running, completed, failed
    queue: str = "default"  # Queue name(s)
    lock: str = ""  # JSON string of lockfile for the stage
    deps_lock: str = ""  # JSON string of lockfile for the dependencies
    deps_hash: str = ""  # Hash of the dependencies

    # Relationships
    parents: List["Job"] = Relationship(
        link_model=JobDependency,
        back_populates="children",
        sa_relationship_kwargs={
            "primaryjoin": "Job.id==JobDependency.child_id",
            "secondaryjoin": "Job.id==JobDependency.parent_id",
        },
    )
    children: List["Job"] = Relationship(
        link_model=JobDependency,
        back_populates="parents",
        sa_relationship_kwargs={
            "primaryjoin": "Job.id==JobDependency.parent_id",
            "secondaryjoin": "Job.id==JobDependency.child_id",
        },
    )


def save_graph_to_db(graph: nx.DiGraph, queues: dict[str, str], changed: list[str]):
    engine = create_engine("sqlite:///jobs.db")
    SQLModel.metadata.create_all(engine)
    with Session(engine) as session:
        for node in nx.topological_sort(graph):
            queue = "default"
            # use fnmatch to match the node name with the custom queues
            for pattern, q in queues.items():
                if fnmatch.fnmatch(node.name, pattern):
                    queue = q
                    break
            job = Job(
                cmd=node.cmd,
                name=node.name,
                queue=queue,
                status="pending" if node.name in changed else "completed",
            )
            session.add(job)
            # add dependencies
            for parent in graph.predecessors(node):
                parent_job = session.exec(
                    select(Job).where(Job.cmd == parent.cmd)
                ).one()
                session.add(JobDependency(parent_id=parent_job.id, child_id=job.id))

        session.commit()


def db_to_graph(db: str = "sqlite:///jobs.db") -> nx.DiGraph:
    engine = create_engine(db)
    with Session(engine) as session:
        jobs = session.exec(select(Job)).all()
        graph = nx.DiGraph()
        for job in jobs:
            graph.add_node(
                job.id,
                **{
                    "name": job.name,
                    "cmd": job.cmd,
                    "status": job.status,
                    "queue": job.queue,
                    "lock": json.loads(job.lock) if job.lock else None,
                    "deps_lock": json.loads(job.deps_lock) if job.deps_lock else None,
                    "deps_hash": job.deps_hash,
                    "group": get_group(job.name),
                },
            )
            for parent in job.parents:
                graph.add_edge(parent.id, job.id)

    return graph


def get_group(name: str) -> list[str]:
    """Extract the group from the job name."""
    parts = name.split("_")
    # check if parts[-1] is a number
    if parts[-1].isdigit():
        return parts[:-2]
    return parts[:-1]


def get_job(db: str = "sqlite:///jobs.db", queues: list | None = None) -> dict | None:
    """
    Get the next job where status is 'pending' and all parents are 'completed'.
    """
    engine = create_engine(db)
    with Session(engine) as session:
        # Fetch jobs with 'pending' status, eagerly loading their parents
        statement = select(Job).where(Job.status == "pending")
        if queues:
            statement = statement.where(Job.queue.in_(queues))
        results = session.exec(statement)

        # Process each job to check if all parents are completed
        for job in results:
            print(len(job.parents))
            if all(parent.status == "completed" for parent in job.parents):
                job.status = "running"
                session.add(job)
                session.commit()
                return {
                    "id": job.id,
                    "name": job.name,
                    "cmd": job.cmd,
                    "queue": job.queue,
                    "status": job.status,
                }
    return None


def set_job_deps_lock(job_id: int, lock: dict, db: str = "sqlite:///jobs.db"):
    engine = create_engine(db)

    reduced_lock = {}

    if x := lock.get("cmd"):
        reduced_lock["cmd"] = x
    if x := lock.get("params"):
        reduced_lock["params"] = x
    if x := lock.get("deps"):
        reduced_lock["deps"] = x

    with Session(engine) as session:
        statement = select(Job).where(Job.id == job_id)
        results = session.exec(statement)
        job = results.one()
        job.deps_lock = json.dumps(reduced_lock)
        job.deps_hash = _get_cache_hash(reduced_lock)
        session.add(job)
        session.commit()


def complete_job(
    job_id: int, lock: dict, db: str = "sqlite:///jobs.db", status: str = "completed"
):
    engine = create_engine(db)
    with Session(engine) as session:
        statement = select(Job).where(Job.id == job_id)
        results = session.exec(statement)
        job = results.one()
        job.status = status
        job.lock = json.dumps(lock)
        session.add(job)
        session.commit()


def get_nodes_and_edges(db: str = "sqlite:///jobs.db") -> tuple[list, list]:
    engine = create_engine(db)
    with Session(engine) as session:
        jobs = session.exec(select(Job)).all()
        nodes = []
        edges = []
        for job in jobs:
            nodes.append(
                {
                    "id": str(job.id),
                    "label": job.name,
                    "status": job.status,
                    "queue": job.queue,
                    "lock": json.loads(job.lock) if job.lock else None,
                    "deps_lock": json.loads(job.deps_lock) if job.deps_lock else None,
                    "deps_hash": job.deps_hash,
                    "group": get_group(job.name),
                }
            )
            for parent in job.parents:
                edges.append({"source": str(parent.id), "target": str(job.id)})
        return nodes, edges
